// This is Array.h


#ifndef ARRAY_H
#define ARRAY_H




template <class Type>
class Array {
	public:
		Array(int numItems=30);
		Array(const Array<Type>& data);
		~Array();


		inline int getNumberOfElements()     { return _maxItemNumberReferenced; }

		Array<Type>& operator=(const Array<Type>& data);
		inline Type& operator[](int index);


	protected:
		int _numItems;
		int _maxItemNumberReferenced;

		Type *_data;

		void _reallocateData(int numItems);


	private:
};







#include <stdlib.h>



//
// The whole impelementation of the linked list class has to be in
// this header file to satisfy the Microsoft compiler.  Don't you 
// just love Microsoft?
//


#ifndef MAX
#define MAX(a, b) ((a)>(b)?(a):(b))
#endif

#ifndef MIN
#define MIN(a, b) ((a)<(b)?(a):(b))
#endif



template<class Type>
Array<Type>::Array(int numItems)
	: _numItems(0),
	  _maxItemNumberReferenced(0),
	  _data(NULL) {
	_reallocateData(numItems);
}




template<class Type>
Array<Type>::Array(const Array& data)
	: _numItems(0),
	  _maxItemNumberReferenced(0),
	  _data(NULL) {
	*this = data;
}




template<class Type>
Array<Type>::~Array() {
	if (_data) {
		delete [] _data;
	}
}







template<class Type>
Array<Type>& 
Array<Type>::operator=(const Array& data) {
	_reallocateData(data._numItems);

	if (_numItems > 0) {
		for (int i=0; i < _numItems; i++) {
			_data[i] = data._data[i];
		}
	}

	_maxItemNumberReferenced = data._maxItemNumberReferenced;

	_sendEvent();

	return *this;
}





template<class Type>
inline Type& Array<Type>::operator[](int index) {
	if (index >= _numItems) {
		_reallocateData(MIN((index+1)*2, index+1000));
	}

	if (index >= _maxItemNumberReferenced) {
		_maxItemNumberReferenced = index+1;
	}

	return _data[index];
}






template<class Type>
void Array<Type>::_reallocateData(int numItems) {
	Type *oldData = _data;
	int oldNumItems = _numItems;

	_numItems = numItems;

	if (_numItems > 0) {
		_data = new Type[_numItems];
	}
	else {
		_data = NULL;
	}


	int endOfLoop = MIN(oldNumItems, _numItems);

	for (int i=0; i < endOfLoop; i++) {
		_data[i] = oldData[i];
	}


	if (oldData) {
		delete [] oldData;
	}
}





#endif    // ARRAY_H


