// This is Image.C
//
// Copyright 1997, 1998     Mark Manca
//

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <string.h>

#include "Image.h"

#ifdef _DEBUG
#include <iostream.h>
#endif



#ifndef MIN
#define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif





// constructors, etc;

Image::Image()
	: _width(0),
	  _height(0),

	  _xScale(1.0),
	  _yScale(1.0),

	  _imageName(NULL),

	  _imageType(GL_RGBA),
	  _imageDataType(GL_UNSIGNED_BYTE),

	  _numChannels(0),
	  _extraRowWidth(0),
	  _imageData(NULL),

	  _colormapSize(0),
	  _colormapDataType(GL_UNSIGNED_BYTE),
	  _colormapData(NULL) {

	  // empty
}


Image::~Image() {
	if (_imageName) {
		delete [] _imageName;
		_imageName = NULL;
	}
	if (_imageData) {
		delete [] _imageData;
		_imageData = NULL;
	}
	if (_colormapData) {
		delete [] _colormapData;
		_colormapData = NULL;
	}
}




GLenum Image::getImageType() {
	return _imageType;
}


GLenum Image::getImageDataType() {
	return _imageDataType;
}


int Image::getExtraBytesPerRow() {
	return _extraRowWidth; 
}



void* Image::getImage() {
	return _imageData;
}



int Image::getNumColorMapEntries() {
	return _colormapSize;
}



GLenum Image::getColormapDataType() {
	return _colormapDataType;
}



void* Image::getColorMap() {
	return _colormapData;
}







Image* Image::getHue() {
	Image* image = new Image;
	int i, j;


	image->_width = _width;
	image->_height = _height;
	image->_xScale = _xScale;
	image->_yScale = _yScale;

	if (_imageName) {
		#ifdef WIN32
    		image->_imageName = _strdup(_imageName);
		#else
			image->_imageName = strdup(_imageName);
		#endif
	}
	else {
		image->_imageName = NULL;
	}

	image->_imageType = GL_COLOR_INDEX;
	image->_numChannels = 1;
	image->_numChannels = 1;


	double max, min, delta, hue;

	if (_imageDataType == GL_UNSIGNED_BYTE) {
		image->_imageDataType = GL_UNSIGNED_BYTE;
		image->_colormapSize = 256;
		image->_colormapDataType = GL_UNSIGNED_BYTE;

		image->_allocateImageData();
		image->_allocateColormapData();

		GLubyte *charPtr = ((GLubyte*)(image->_colormapData));

		for (i=0; i < image->_colormapSize; i++) {
			*charPtr++ = (GLubyte)i;
			*charPtr++ = (GLubyte)i;
			*charPtr++ = (GLubyte)i;
		}


		char redChar, greenChar, blueChar;

		GLubyte *thisImageCharPtr = ((GLubyte*)_imageData);
		charPtr = ((GLubyte*)(image->_imageData));

		if (_imageType == GL_RGB) {
			for (i = _height; i > 0; i--) {
				for (j=_width; j > 0; j--) {
					redChar = *thisImageCharPtr++;
					greenChar = *thisImageCharPtr++;
					blueChar = *thisImageCharPtr++;

					max = MAX(MAX(redChar, greenChar), blueChar);
					min = MIN(MIN(redChar, greenChar), blueChar);
					delta = max - min;

					if (redChar == max) {
						hue = ((double)(greenChar-blueChar))/delta;
					}
					else if (greenChar == max) {
						hue = 2.0 + ((double)(blueChar-redChar))/delta;
					}
					else if (blueChar == max) {
						hue = 4.0 + ((double)(redChar-greenChar))/delta;
					}

					hue *= 60;
					if (hue < 0.0) {
						hue += 360;
					}

					*charPtr++ = (char)(255.0 * hue / 360.0);
				}
				thisImageCharPtr += _extraRowWidth;
				charPtr += image->_extraRowWidth;
			}
		}
		else if (_imageType == GL_RGBA) {
			for (i = _height; i > 0; i--) {
				for (j=_width; j > 0; j--) {
					redChar = *thisImageCharPtr++;
					greenChar = *thisImageCharPtr++;
					blueChar = *thisImageCharPtr++;

					max = MAX(MAX(redChar, greenChar), blueChar);
					min = MIN(MIN(redChar, greenChar), blueChar);
					delta = max - min;

					if (redChar == max) {
						hue = ((double)(greenChar-blueChar))/delta;
					}
					else if (greenChar == max) {
						hue = 2.0 + ((double)(blueChar-redChar))/delta;
					}
					else if (blueChar == max) {
						hue = 4.0 + ((double)(redChar-greenChar))/delta;
					}

					hue *= 60;
					if (hue < 0.0) {
						hue += 360;
					}

					*charPtr++ = (char)(255.0 * hue / 360.0);

					thisImageCharPtr++;
				}
				thisImageCharPtr += _extraRowWidth;
				charPtr += image->_extraRowWidth;
			}
		}
		else if (_imageType == GL_COLOR_INDEX) {
			GLubyte* colormapCharPtr = ((GLubyte*)_colormapData);
			GLubyte* colorCharPtr;

			for (i = _height; i > 0; i--) {
				for (j=_width; j > 0; j--) {
					colorCharPtr = &colormapCharPtr[*thisImageCharPtr++];

					redChar = *colorCharPtr++;
					greenChar = *colorCharPtr++;
					blueChar = *colorCharPtr++;

					max = MAX(MAX(redChar, greenChar), blueChar);
					min = MIN(MIN(redChar, greenChar), blueChar);
					delta = max - min;

					if (redChar == max) {
                        hue = ((double)(greenChar-blueChar))/delta;
					}
					else if (greenChar == max) {
                        hue = 2.0 + ((double)(blueChar-redChar))/delta;
					}
					else if (blueChar == max) {
                        hue = 4.0 + ((double)(redChar-greenChar))/delta;
					}

					hue *= 60;
					if (hue < 0.0) {
                        hue += 360;
					}

					*charPtr++ = (char)(255.0 * hue / 360.0);
				}
				thisImageCharPtr += _extraRowWidth;
				charPtr += image->_extraRowWidth;
			}
		}
	}


	return image;
}







int Image::getDepth() {
	int numBitsPerChannel;

	switch (_imageDataType) {
		case GL_UNSIGNED_BYTE:
		case GL_BYTE:
			numBitsPerChannel = 8;
			break;

		case GL_UNSIGNED_SHORT:
		case GL_SHORT:
			numBitsPerChannel = 16;
			break;

		case GL_INT:
		case GL_UNSIGNED_INT:
		case GL_FLOAT:
			numBitsPerChannel = 32;
			break;

		default:
			numBitsPerChannel = 0;
			break;
	}

	return _numChannels * numBitsPerChannel;
}




int Image::getNumChannels() {
	return _numChannels;
}




void Image::setScale(float scale) {
	_xScale = scale;
	_yScale = scale;
}

void Image::setXScale(float xScale) {
	_xScale = xScale;
}

void Image::setYScale(float yScale) {
	_yScale = yScale;
}



void Image::setSize(int width, int height) {
	_width = width;
	_height = height;

	// more work needs to be done here.
	_numChannels = 4;

	if( _imageData != NULL )
		{
		delete [] _imageData;
		_imageData = NULL;
		}

	_allocateImageData();
}




float Image::getXScale() {
	return _xScale;
}

float Image::getYScale() {
	return _yScale;
}



void Image::getSize(int *width, int *height) {
	*width = _width;
	*height = _height;
}

int Image::getWidth() {
	return _width;
}

int Image::getHeight() {
	return _height;
}



void Image::draw() {
	if (!_imageData) {
		return;
	}

	glPushAttrib(GL_PIXEL_MODE_BIT | GL_CURRENT_BIT);
	glPixelZoom(_xScale, _yScale);
	glRasterPos2i(0, 0);
	glDrawPixels(_width, _height, _imageType, _imageDataType, _imageData);
	glPopAttrib();
}


void Image::drawRow(int row) {
	draw(0, row, _width, 1);
}


void Image::draw(int x, int y, int width, int height) {
	int skipRows, skipPixels, rowLength;

	glGetIntegerv(GL_UNPACK_SKIP_ROWS, &skipRows);
	glGetIntegerv(GL_UNPACK_SKIP_PIXELS, &skipPixels);
	glGetIntegerv(GL_UNPACK_ROW_LENGTH, &rowLength);

	glPixelStorei(GL_UNPACK_SKIP_ROWS, y);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, x);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, _width);


	glPushAttrib(GL_PIXEL_MODE_BIT | GL_CURRENT_BIT);
	glPixelZoom(_xScale, _yScale);
	glRasterPos2i(x, y);
	glDrawPixels(width, height, _imageType, _imageDataType, _imageData);
	glPopAttrib();


	glPixelStorei(GL_UNPACK_SKIP_ROWS, skipRows);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, skipPixels);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, rowLength);
}




void Image::drawRed() {
	glPushAttrib(GL_PIXEL_MODE_BIT | GL_CURRENT_BIT);
	glPixelTransferf(GL_GREEN_SCALE, 0.0);
	glPixelTransferf(GL_BLUE_SCALE, 0.0);

	glPixelZoom(_xScale, _yScale);
	glRasterPos2i(0, 0);
	glDrawPixels(_width, _height, _imageType, _imageDataType, _imageData);
	glPopAttrib();
}


void Image::drawGreen() {
	glPushAttrib(GL_PIXEL_MODE_BIT | GL_CURRENT_BIT);
	glPixelTransferf(GL_RED_SCALE, 0.0);
	glPixelTransferf(GL_BLUE_SCALE, 0.0);

	glPixelZoom(_xScale, _yScale);
	glRasterPos2i(0, 0);
	glDrawPixels(_width, _height, _imageType, _imageDataType, _imageData);
	glPopAttrib();
}


void Image::drawBlue() {
	glPushAttrib(GL_PIXEL_MODE_BIT | GL_CURRENT_BIT);
	glPixelTransferf(GL_RED_SCALE, 0.0);
	glPixelTransferf(GL_GREEN_SCALE, 0.0);

	glPixelZoom(_xScale, _yScale);
	glRasterPos2i(0, 0);
	glDrawPixels(_width, _height, _imageType, _imageDataType, _imageData);
	glPopAttrib();
}


void Image::drawAlpha() {
	glPushAttrib(GL_PIXEL_MODE_BIT | GL_CURRENT_BIT);
	glPixelZoom(_xScale, _yScale);
	glRasterPos2i(0, 0);

	//
	// haven't done this one yet....
	//

	glPopAttrib();
}






void Image::update() {
	glPushAttrib(GL_PIXEL_MODE_BIT);
    glPixelZoom(_xScale, _yScale);
    glReadPixels(0, 0, _width, _height, 
	             _imageType, _imageDataType, _imageData);
	glPopAttrib();
}






void Image::update(int x, int y, int width, int height) {
    int skipRows, skipPixels, rowLength;

    glGetIntegerv(GL_PACK_SKIP_ROWS, &skipRows);
    glGetIntegerv(GL_PACK_SKIP_PIXELS, &skipPixels);
    glGetIntegerv(GL_PACK_ROW_LENGTH, &rowLength);

    glPixelStorei(GL_PACK_SKIP_ROWS, y);
    glPixelStorei(GL_PACK_SKIP_PIXELS, x);
    glPixelStorei(GL_PACK_ROW_LENGTH, _width);


	glPushAttrib(GL_PIXEL_MODE_BIT);
    glPixelZoom(_xScale, _yScale);
    glReadPixels(x, y, width, height, _imageType, _imageDataType, _imageData);
	glPopAttrib();


    glPixelStorei(GL_PACK_SKIP_ROWS, skipRows);
    glPixelStorei(GL_PACK_SKIP_PIXELS, skipPixels);
    glPixelStorei(GL_PACK_ROW_LENGTH, rowLength);
}







int Image::setName(const char *name) {

	if (_imageName) {
		delete [] _imageName;
		_imageName = NULL;
	}

#ifdef WIN32
	_imageName = _strdup(name);
#else
	_imageName = strdup(name);
#endif

	if (_imageName) {
		return 1;
	}
	else {
		return 0;
	}
}






const char* Image::getName() {
	return _imageName;
}






//
// This function will only work if this is an RGBA image.
//
Color& Image::operator[](int pixelOffset) {
	int numRows = pixelOffset / _width;
	int numCols = pixelOffset % _width;
	int rowDataSize = _width * _numChannels + _extraRowWidth;
	int offset = rowDataSize * numRows + numCols * _numChannels;

	if (offset >= ((_width*_numChannels) + _extraRowWidth) * _height) {
		//
		// Lets core dump shall we....
		//
		char *hello = NULL;
		cerr << "DOH!!!!!!!  You've indexed an image out of bounds." << endl;
		*hello = 'c';
	}

	return *((Color*)((char*)_imageData + offset));
}









void Image::_allocateImageData() {
	int imageRowDataSize;
	int remainder;
	int allocationSize;


	//
	// In some implementations of OpenGL (namely SGIs) you must pad images
	// by a certain number of bytes to word align rows of the image for
	// efficiency. Although it is possible to set up OpenGL to eliminate
	// this need, we pad our image becase it makes drawing to the screen
	// faster.
	//
	glGetIntegerv(GL_UNPACK_ALIGNMENT, &_extraRowWidth);


	//
	// COME BACK HERE
	//
	// I shouldn't do this....
	//
	//_extraRowWidth = 4;
	_extraRowWidth = 0;


	if (_imageDataType == GL_BYTE || _imageDataType == GL_UNSIGNED_BYTE) {
		//
		// the (height*(width%4)) is there to handle how glDrawPixels
		// works. glDrawPixels starts drawing the next row at the next
		// word boundry, not the next byte boundry as you would think.
		// So we must make sure that each row is ends on a word boundry
		//
		imageRowDataSize = _width*_numChannels*sizeof(char);

		_extraRowWidth = (_extraRowWidth == 1 ? 0 : _extraRowWidth);

		if (_extraRowWidth) {
			remainder = imageRowDataSize % _extraRowWidth;
			_extraRowWidth = (remainder ? _extraRowWidth - remainder : 0);
		}

		imageRowDataSize += _extraRowWidth;

		allocationSize = ((_width*_numChannels)+_extraRowWidth)*_height;

		if (_imageDataType == GL_BYTE) {
			_imageData = new GLbyte[allocationSize];
		}
		else {        // _imageDataType == GL_UNSIGNED_BYTE
			_imageData = new GLubyte[allocationSize];
		}
	}
	else if (_imageDataType == GL_SHORT || 
	         _imageDataType == GL_UNSIGNED_SHORT) {
		imageRowDataSize = _width*_numChannels*sizeof(short);

		_extraRowWidth = (_extraRowWidth == 1 ? 0 : _extraRowWidth);

		if (_extraRowWidth) {
			remainder = imageRowDataSize % _extraRowWidth;
			_extraRowWidth = (remainder ? _extraRowWidth - remainder : 0);
		}

		allocationSize = ((_width*_numChannels)+_extraRowWidth)*_height;

		if (_imageDataType == GL_SHORT) {
			_imageData = new GLshort[allocationSize];
		}
		else {        // _imageDataType == GL_UNSIGNED_SHORT
			_imageData = new GLushort[allocationSize];
		}
	}
	else if (_imageDataType == GL_INT || _imageDataType == GL_UNSIGNED_INT) {
		imageRowDataSize = _width*_numChannels*sizeof(int);

		_extraRowWidth = (_extraRowWidth == 1 ? 0 : _extraRowWidth);

		if (_extraRowWidth) {
			remainder = imageRowDataSize % _extraRowWidth;
			_extraRowWidth = (remainder ? _extraRowWidth - remainder : 0);
		}

		allocationSize = ((_width*_numChannels)+_extraRowWidth)*_height;

		if (_imageDataType == GL_INT) {
			_imageData = new GLint[allocationSize];
		}
		else {        // _imageDataType == GL_UNSIGNED_INT
            _imageData = new GLuint[allocationSize];
		}
	}
	else if (_imageDataType == GL_FLOAT) {
		imageRowDataSize = _width*_numChannels*sizeof(float);

		_extraRowWidth = (_extraRowWidth == 1 ? 0 : _extraRowWidth);

		if (_extraRowWidth) {
			remainder = imageRowDataSize % _extraRowWidth;
			_extraRowWidth = (remainder ? _extraRowWidth - remainder : 0);
		}

		allocationSize = ((_width*_numChannels)+_extraRowWidth)*_height;

		_imageData = new GLfloat[allocationSize];
    }
}






void Image::_allocateColormapData() {
	if (_colormapData) {
		delete [] _colormapData;
		_colormapData = NULL;
	}


	if (_colormapDataType == GL_BYTE && _colormapSize) {
		_colormapData = new GLbyte[_colormapSize * 3];
	}
	else if (_colormapDataType == GL_UNSIGNED_BYTE && _colormapSize) {
		_colormapData = new GLubyte[_colormapSize * 3];
	}
	else if (_colormapDataType == GL_SHORT && _colormapSize) {
		_colormapData = new GLshort[_colormapSize * 3];
	}
	else if (_colormapDataType == GL_UNSIGNED_SHORT && _colormapSize) {
		_colormapData = new GLushort[_colormapSize * 3];
	}
	else if (_colormapDataType == GL_INT && _colormapSize) {
		_colormapData = new GLint[_colormapSize * 3];
	}
	else if (_colormapDataType == GL_UNSIGNED_INT && _colormapSize) {
		_colormapData = new GLuint[_colormapSize * 3];
	}
	else if (_colormapDataType == GL_FLOAT && _colormapSize) {
		_colormapData = new GLfloat[_colormapSize * 3];
	}
	else {
		_colormapData = NULL;
	}
}





