// This is ImageFileBMPLoad.cpp
//
// Copyright 1997, 1998     Mark Manca
//

// This file contains the code to load a bmp image

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <string.h>
#include <stdio.h>

#include "ImageFile.h"
#include "bmpFileHeader.h"

#ifdef BIG_ENDIAN
#include "byteSwap.h"
#endif



// These functions assumes that the file to be read is already open

int ImageFile::_readBMPHeader() {
	bmpFileHeader fileHeader;
	bmpInfoHeader infoHeader;

	return _readBMPHeader(&fileHeader, &infoHeader);
}




int ImageFile::_readBMPHeader(bmpFileHeader *fileHeader, 
							  bmpInfoHeader* infoHeader) {
	int status;


	if (!__imageFile || !__isOpen) {
    	return 0;
	}

	//
	// get to the beginning of the file so we can read the 
	// header into the struct
	//
	status = fseek(__imageFile, 0, SEEK_SET);
	if (status) {
		return 0;			// an error occured
	}


	//
	// read in the file header into the data structure
	//
	status = fread((char *)fileHeader, 1, sizeof(bmpFileHeader), __imageFile);

	// should we use ferror here??
	if (status != sizeof(bmpFileHeader)) {
		return 0;
	}



	//
	// read in the information header so we can figure out more things about
	// the image we are reading in.
	//
	status = fread((char *)infoHeader, 1, sizeof(bmpInfoHeader), __imageFile);

	// should we use ferror here??
	if (status != sizeof(bmpInfoHeader)) {
		return 0;
	}



	//
	// since the bmp file format was designed for use on WinTel platforms, we
	// are going to have to swap some things around on machines that have
	// big endian word order.
	//
	#ifdef BIG_ENDIAN
		fileHeader->sizeOfFile = halfwordAndByteSwap(fileHeader->sizeOfFile);
		fileHeader->reserved1 = byteSwap(fileHeader->reserved1);
		fileHeader->reserved2 = byteSwap(fileHeader->reserved2);
		fileHeader->bitmapOffset = halfwordAndByteSwap(fileHeader->bitmapOffset);

		infoHeader->sizeOfStruct = halfwordAndByteSwap(infoHeader->sizeOfStruct);
		infoHeader->width = halfwordAndByteSwap(infoHeader->width);
		infoHeader->height = halfwordAndByteSwap(infoHeader->height);
		infoHeader->planes = byteSwap(infoHeader->planes);
		infoHeader->bitsPerPixel = byteSwap(infoHeader->bitsPerPixel);
		infoHeader->compression = halfwordAndByteSwap(infoHeader->compression);
		infoHeader->size = halfwordAndByteSwap(infoHeader->size);
		infoHeader->xPixelsPerMeter = halfwordAndByteSwap(infoHeader->xPixelsPerMeter);
		infoHeader->yPixelsPerMeter = halfwordAndByteSwap(infoHeader->yPixelsPerMeter);
		infoHeader->indicesUsed = halfwordAndByteSwap(infoHeader->indicesUsed);
		infoHeader->importantIndices = halfwordAndByteSwap(infoHeader->importantIndices);
	#endif



	//
	// set up some member variables to correspond to the values that we just
	// read in from the file.
	//
	_width = infoHeader->width;
	_height = infoHeader->height;

	_imageType;
	_imageDataType;

	_numChannels;

	_colormapSize = infoHeader->indicesUsed;
	_colormapDataType = GL_UNSIGNED_BYTE;


#if 0
	The biBitCount member of the BITMAPINFOHEADER structure determines the number of bits that define each pixel and the maximum number of colors in the bitmap. This member can be one of the following values: 


Value 



Meaning 

1 



The bitmap is monochrome, and the bmiColors member contains two entries. Each bit in the bitmap array represents a pixel. If the bit is clear, the pixel is displayed with the color of the first entry in the bmiColors table; if the bit is set, the pixel has the color of the second entry in the table. 

4 



The bitmap has a maximum of 16 colors, and the bmiColors member contains up to 16 entries. Each pixel in the bitmap is represented by a 4-bit index into the color table. For example, if the first byte in the bitmap is 0x1F, the byte represents two pixels. The first pixel contains the color in the second table entry, and the second pixel contains the color in the sixteenth table entry. 

8 



The bitmap has a maximum of 256 colors, and the bmiColors member contains up to 256 entries. In this case, each byte in the array represents a single pixel. 

16 



The bitmap has a maximum of 2^16 colors. If the biCompression member of the BITMAPINFOHEADER is BI_RGB, the bmiColors member is NULL. Each WORD in the bitmap array represents a single pixel. The relative intensities of red, green, and blue are represented with 5 bits for each color component. The value for blue is in the least significant 5 bits, followed by 5 bits each for green and red, respectively. The most significant bit is not used. 

If the biCompression member of the BITMAPINFOHEADER is BI_BITFIELDS, the bmiColors member contains three DWORD color masks that specify the red, green, and blue components, respectively, of each pixel. Each WORD in the bitmap array represents a single pixel. 

Windows NT: When the biCompression member is BI_BITFIELDS, bits set in each DWORD mask must be contiguous and should not overlap the bits of another mask. All the bits in the pixel do not have to be used. 

Windows 95: When the biCompression member is BI_BITFIELDS, Windows 95 supports only the following 16bpp color masks: A 5-5-5 16-bit image, where the blue mask is 0x001F, the green mask is 0x03E0, and the red mask is 0x7C00; and a 5-6-5 16-bit image, where the blue mask is 0x001F, the green mask is 0x07E0, and the red mask is 0xF800. 

24 



The bitmap has a maximum of 2^24 colors, and the bmiColors member is NULL. Each 3-byte triplet in the bitmap array represents the relative intensities of blue, green, and red, respectively, for a pixel. 

32 



The bitmap has a maximum of 2^32 colors. If the biCompression member of the BITMAPINFOHEADER is BI_RGB, the bmiColors member is NULL. Each DWORD in the bitmap array represents the relative intensities of blue, green, and red, respectively, for a pixel. The high byte in each DWORD is not used. 

If the biCompression member of the BITMAPINFOHEADER is BI_BITFIELDS, the bmiColors member contains three DWORD color masks that specify the red, green, and blue components, respectively, of each pixel. Each DWORD in the bitmap array represents a single pixel. 

Windows NT: When the biCompression member is BI_BITFIELDS, bits set in each DWORD mask must be contiguous and should not overlap the bits of another mask. All the bits in the pixel do not have to be used. 

Windows 95: When the biCompression member is BI_BITFIELDS, Windows 95 supports only the following 32bpp color mask: The blue mask is 0x000000FF, the green mask is 0x0000FF00, and the red mask is 0x00FF0000. 



The biClrUsed member of the BITMAPINFOHEADER structure specifies the number of color indices in the color table that are actually used by the bitmap. If the biClrUsed member is set to zero, the bitmap uses the maximum number of colors corresponding to the value of the biBitCount member. 

The colors in the bmiColors table should appear in order of importance. 


	_imageType = GL_COLOR_INDEX;
	_imageDataType = GL_UNSIGNED_BYTE;
	_numChannels = 1;
#endif


	#ifdef _DEBUG
		printf("width: %d     height: %d\n", _width, _height);
		printf("offsetToData: %d\n", fileHeader->bitmapOffset);
		printf("bitsPerPixel: %d\n", infoHeader->bitsPerPixel);
		printf("compressionType: %d\n", infoHeader->compression);
		printf("xPixelsPerMeter: %d\n", infoHeader->xPixelsPerMeter);
		printf("yPixelsPerMeter: %d\n", infoHeader->yPixelsPerMeter);
		printf("indicesUsed: %d\n", infoHeader->indicesUsed);
		printf("importantIndices: %d\n", infoHeader->importantIndicies);
		fflush(stdout);
	#endif


	//
	// if we've made it this far, return 1
	//
	return 1;
}








int ImageFile::_readBMPFile() {
	bmpFileHeader fileHeader;
	bmpInfoHeader infoHeader;

	if (!_readBMPHeader(&fileHeader, &infoHeader)) {
		return 0;
	}


	//
	// allocate the memory that will be used to store the image
	//
	_allocateImageData();


	//
	// allocate the color map that will be necessary to use since
	// all gif images are color mapped.
	//
	_allocateColormapData();


	return 1;
}

