/******************************************************************
 * Raytrace.c 
 *
 * The ray tracing subroutines 
 *
 *         Program #1 -- basic ray tracer
 ******************************************************************/

#include <stdio.h>
#include <math.h>
#include "Vector3f.h"
#include "Matrix4f.h"
#include "SceneIO.h"
#include "Intersect.h"
#include "Raytrace.h"

#ifndef min
#define min(a,b) ((a)<(b)?(a):(b))
#endif

/**************************************************************
 * TraceRay()
 *
 * Traces a ray to the first intersection with the scene
 * and shades it.
 **************************************************************/
void TraceRay (SceneIO *scene, int level, float weight, Ray *ray, Vector3f color, 
			   RTParameters* g_params)
{
	Isect I;
	Point3f P, N;

	/* Find the closest intersection with the scene */
	if ( Intersect(ray, scene, &I, g_params) )
	{
		/* calc intersection point */
		RayPoint (ray, I.t, P);

		/* get the surface normal */
		if ( I.object->type == SPHERE_OBJ)
		{
			SphereNormal(I.object, P, N);
		}

		if ( I.object->type == POLYSET_OBJ)
		{
			Vector3f_copy(I.normal, N);
		}

		/* reverse the normal if necesary */
		if (Vector3f_dot(ray->D, N) > 0.)
		{
			Vector3f_scale (-1.0, N);
		}

		/* color the pixel */
		Shade (scene, level, weight, P, N, ray->D, &I, color, g_params);
	}
	else
	{
		ShadeBackground (ray, color);
	}
}

/**************************************************************
 * Shade()
 *
 * Finds the color for a given intersection point.
 * Uses the Whitted model, and recursively calls Trace for
 * the reflected and transmitted rays.
 **************************************************************/
void Shade(SceneIO *scene, int level, float weight, Vector3f P, Vector3f N, Vector3f I,
			struct IsectType *intersect, Vector3f color,  RTParameters* g_params)
{
     MaterialIO *surf;

     surf = intersect->object->material;

     Vector3f_set (color, 0., 0., 0.);

     /* add ambient term, Ia is assumed to be (1,1,1) */ 
	 Vector3f_copy(surf->ambColor, color);

     /*  You will need to fill this in */

     ClampColor(color);

}


/**************************************************************
 * ShadeBackground()
 *
 * Shade a pixel with the background color, black.
 **************************************************************/
void ShadeBackground (Ray *ray, Vector3f color)
{
	Vector3f_set (color, .0, .0, .0);
}








