#include <windows.h>
#include <GL/gl.h>

#include "RenderQuadMesh.h"

/* return the cross product c = a cross b */
static void cross(float a[3], float b[3], float c[3])
{
	c[0] = (a[1]*b[2]) - (a[2]*b[1]);
	c[1] = (a[2]*b[0]) - (a[0]*b[2]);
	c[2] = (a[0]*b[1]) - (a[1]*b[0]);
}

// c = a - b
static void sub(float a[3], float b[3], float c[3])
{
	c[0] = a[0]-b[0];
	c[1] = a[1]-b[1];
	c[2] = a[2]-b[2];
}


//-------------------------------------------------------------------
//-------------------------------------------------------------------
void RenderQuadMesh_POM_PVN( PolySetIO *pset, MaterialIO *mat, long num_mat )
{
		//
	//	Init Per Object Material
	//
	float color[4];

	color[3] = 1.0f - mat->ktran;
				
	color[0] = mat->ambColor[0];
	color[1] = mat->ambColor[1];
	color[2] = mat->ambColor[2];

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, color);

	color[0] = mat->diffColor[0];
	color[1] = mat->diffColor[1];
	color[2] = mat->diffColor[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, color);

	color[0] = mat->specColor[0];
	color[1] = mat->specColor[1];
	color[2] = mat->specColor[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, color);

	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, mat->shininess);

	color[0] = mat->emissColor[0];
	color[1] = mat->emissColor[1];
	color[2] = mat->emissColor[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, color);



	//
	//	Draw Quads
	//

	fprintf(stderr,"Rendering Quad Mesh - Per Object Material - Per Vertex Normals \n");

	//	Loop over quadrilaterals (faces)
	glBegin( GL_QUADS );
	for( int i=0 ; i<pset->numPolys ; i++ )
	{
		glNormal3fv( pset->poly[i].vert[0].norm );
		glVertex3fv( pset->poly[i].vert[0].pos );

		glNormal3fv( pset->poly[i].vert[1].norm );
		glVertex3fv( pset->poly[i].vert[1].pos );

		glNormal3fv( pset->poly[i].vert[2].norm );
		glVertex3fv( pset->poly[i].vert[2].pos );

		glNormal3fv( pset->poly[i].vert[3].norm );
		glVertex3fv( pset->poly[i].vert[3].pos );
	}
	glEnd();

}








void RenderQuadMesh_POM_PFN( PolySetIO *pset, MaterialIO *mat, long num_mat )
{
	//
	//	Init Per Object Material
	//
	float color[4];

	color[3] = 1.0f - mat->ktran;
				
	color[0] = mat->ambColor[0];
	color[1] = mat->ambColor[1];
	color[2] = mat->ambColor[2];

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, color);

	color[0] = mat->diffColor[0];
	color[1] = mat->diffColor[1];
	color[2] = mat->diffColor[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, color);

	color[0] = mat->specColor[0];
	color[1] = mat->specColor[1];
	color[2] = mat->specColor[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, color);

	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, mat->shininess);

	color[0] = mat->emissColor[0];
	color[1] = mat->emissColor[1];
	color[2] = mat->emissColor[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, color);



	//
	//	Draw Quads
	//

	fprintf(stderr,"Rendering Quad Mesh - Per Object Material - Per Face Normals \n");

	//	Loop over quadrilaterals (faces)
	glBegin( GL_QUADS );
	float n[3],v1[3],v2[3];
	for( int i=0 ; i<pset->numPolys ; i++ )
	{
		//
		// compute a normal, assume that the quadrilateral is planer, so 
		// just use the first three vertexes.
		//
		sub( pset->poly[i].vert[0].pos, pset->poly[i].vert[1].pos, v1 );
		sub( pset->poly[i].vert[2].pos, pset->poly[i].vert[1].pos, v2 );
		cross( v2,v1,n);
		glNormal3fv( n );

		glVertex3fv( pset->poly[i].vert[0].pos );
		glVertex3fv( pset->poly[i].vert[1].pos );
		glVertex3fv( pset->poly[i].vert[2].pos );
		glVertex3fv( pset->poly[i].vert[3].pos );
	}
	glEnd();

}








void RenderQuadMesh_PVM_PVN( PolySetIO *pset, MaterialIO *mat, long num_mat )
{
	//
	//	Init Per Object Material
	//
	float color[4];

	color[3] = 1.0f - mat->ktran;
				
	color[0] = mat->ambColor[0];
	color[1] = mat->ambColor[1];
	color[2] = mat->ambColor[2];

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, color);

	color[0] = mat->diffColor[0];
	color[1] = mat->diffColor[1];
	color[2] = mat->diffColor[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, color);

	color[0] = mat->specColor[0];
	color[1] = mat->specColor[1];
	color[2] = mat->specColor[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, color);

	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, mat->shininess);

	color[0] = mat->emissColor[0];
	color[1] = mat->emissColor[1];
	color[2] = mat->emissColor[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, color);



	//
	//	Draw Quadrilaterals
	//

	fprintf(stderr,"Rendering Quad Mesh - Per Vertex Material - Per Vertex Normals \n");

	//	Loop over quadrilaterals (faces)
	glBegin( GL_QUADS );
	for( int i=0 ; i<pset->numPolys ; i++ )
	{
		//fprintf(stderr,"%d\n", pset->poly[i].numVertices );
		glNormal3fv( pset->poly[i].vert[0].norm );
		glColor3fv(  mat[pset->poly[i].vert[0].materialIndex].diffColor);
		glVertex3fv( pset->poly[i].vert[0].pos );

		glNormal3fv( pset->poly[i].vert[1].norm );
		glColor3fv(  mat[pset->poly[i].vert[1].materialIndex].diffColor);
		glVertex3fv( pset->poly[i].vert[1].pos );

		glNormal3fv( pset->poly[i].vert[2].norm );
		glColor3fv(  mat[pset->poly[i].vert[2].materialIndex].diffColor);
		glVertex3fv( pset->poly[i].vert[2].pos );

		glNormal3fv( pset->poly[i].vert[3].norm );
		glColor3fv(  mat[pset->poly[i].vert[3].materialIndex].diffColor);
		glVertex3fv( pset->poly[i].vert[3].pos );
	}
	glEnd();
}











void RenderQuadMesh_PVM_PFN( PolySetIO *pset, MaterialIO *mat, long num_mat )
{
	//
	//	Init Per Object Material
	//
	float color[4];

	color[3] = 1.0f - mat->ktran;
				
	color[0] = mat->ambColor[0];
	color[1] = mat->ambColor[1];
	color[2] = mat->ambColor[2];

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, color);

	color[0] = mat->diffColor[0];
	color[1] = mat->diffColor[1];
	color[2] = mat->diffColor[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, color);

	color[0] = mat->specColor[0];
	color[1] = mat->specColor[1];
	color[2] = mat->specColor[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, color);

	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, mat->shininess);

	color[0] = mat->emissColor[0];
	color[1] = mat->emissColor[1];
	color[2] = mat->emissColor[2];
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, color);



	//
	//	Draw Quads
	//

	fprintf(stderr,"Rendering Quad Mesh - Per Vertex Material - Per Face Normals \n");

	//	Loop over quadrilaterals (faces)
	glBegin( GL_QUADS );
	float n[3],v1[3],v2[3];
	for( int i=0 ; i<pset->numPolys ; i++ )
	{
		//	compute a normal
		sub( pset->poly[i].vert[0].pos, pset->poly[i].vert[1].pos, v1 );
		sub( pset->poly[i].vert[2].pos, pset->poly[i].vert[1].pos, v2 );
		cross( v2,v1,n);
		glNormal3fv( n );

		glColor3fv(  mat[pset->poly[i].vert[0].materialIndex].diffColor);
		glVertex3fv( pset->poly[i].vert[0].pos );

		glColor3fv(  mat[pset->poly[i].vert[1].materialIndex].diffColor);
		glVertex3fv( pset->poly[i].vert[1].pos );

		glColor3fv(  mat[pset->poly[i].vert[2].materialIndex].diffColor);
		glVertex3fv( pset->poly[i].vert[2].pos );

		glColor3fv(  mat[pset->poly[i].vert[3].materialIndex].diffColor);
		glVertex3fv( pset->poly[i].vert[3].pos );
	}
	glEnd();
}


